# 后端研发建议


<!-- TOC -->

- [后端研发建议](#后端研发建议)
    - [文档建议](#文档建议)
        - [plantuml 工具](#plantuml-工具)
    - [用于软件架构的 C4 模型](#用于软件架构的-c4-模型)
    - [工程结构建议](#工程结构建议)
    - [框架选型建议](#框架选型建议)
    - [单元测试建议](#单元测试建议)
        - [引入 spock](#引入-spock)
        - [单元测试准则](#单元测试准则)
    - [开发规范](#开发规范)
        - [预防空指针](#预防空指针)
        - [Spring 注入](#spring-注入)
        - [提升代码可阅读性](#提升代码可阅读性)
            - [模型转换使用自定义的 Converter](#模型转换使用自定义的-converter)
            - [spring 的配置注入](#spring-的配置注入)
            - [其他](#其他)
        - [Rest API 接口规范](#rest-api-接口规范)
            - [命名规范](#命名规范)
            - [资源集合 vs 单个资源](#资源集合-vs-单个资源)
            - [避免层级过深的 uri （重要）](#避免层级过深的-uri-重要)
            - [HTTP 方法](#http-方法)
    - [版本与文档规范](#版本与文档规范)
        - [版本管理](#版本管理)
        - [README 规范](#readme-规范)
        - [RELEASE 文档规范](#release-文档规范)
            - [CHANGELOG 格式](#changelog-格式)

<!-- /TOC -->

## 文档建议

使用 markdown + [plantuml](https://plantuml.com/zh/)，可以有以下收益:

* 可以版本控制 uml 的迭代
* 提供 [DDD](https://www.cnblogs.com/netfocus/archive/2011/10/10/2204949.html) 的实践基础准备
* uml 图风格统一


详细见 demo: http://github.miguatech.com/chenlingmin/test/blob/master/README.md

### plantuml 工具

* vscode: 安装 plantuml + Markdown Preview Enhanced
* idea: Languages & Frameworks > Markdown 点击 `Install button next to the PlantUML framework isn't installed` 安装即可.



## 用于软件架构的 C4 模型

https://www.infoq.cn/article/C4-architecture-model

https://github.com/RicardoNiepel/C4-PlantUML




## 工程结构建议


建议使用 project 级别的模块，可以利用构建脚本来约束模块之间的关系。

如下

```plantuml

!define module(x, desc) class "desc" as x << (M,#ffebf3)  >>

left to right direction
skinparam classFontColor #3b0018
skinparam classArrowColor #ff0066
skinparam classBorderColor #ff0066
skinparam classBackgroundColor ##f6f4ee
skinparam shadowing false
'skinparam linetype polyline
skinparam linetype ortho



hide methods
hide stereotypes
'hide field

module(server, "xxx-server") {
    负责提供可执行的服务
}
module(api, "xxx-api") {
    定义 api
}
module(queryapi, "xxx-query-api") {
    定义查询的 api
}
module(controller, "xxx-controller") {
    实现 api 
}
module(querycontroller, "xxx-query-controller") {
    实现查询类api 
}
module(job, "xxx-job") {
    实现任务
}
module(listener, "xxx-listener"){
    实现外部监听
}
module(core, "xxx-core") {
    实现核心业务
}
module(queryservice, "xxx-query-service") {
    实现查询服务
}
module(repository, "xxx-repository") {
    数据存储
}
module(client, "xxx-client") {
    调用外部接口
}


server --> controller
server --> querycontroller
server --> job 
server --> listener
querycontroller --> queryapi
querycontroller --> queryservice 
controller --> api
controller --> core
core --> client
core --> repository
queryservice --> repository
job --> core 
listener --> core

```

## 框架选型建议

依照 [CQRS(查询命令分离)](https://www.cnblogs.com/netfocus/p/4055346.html) 的架构思想，上面的工程结构，拆分出了 query 与 command 的层次，可以在 command 层引入 spring-data-jpa，query 层可以使用 mybatis 或后续的其他非关系型数据的扩展。


```plantuml
!includeurl https://raw.githubusercontent.com/RicardoNiepel/C4-PlantUML/master/C4_Container.puml

Person(command, "命令")
Person(query, "查询")

System_Boundary(interface, "接口") {
  System_Boundary(queryModel, "查询模型") 
  System_Boundary(commandModel, "命令模型") 
  Container(event, "业务事件", "消息队列/或单体应用的事件模型")

}
ContainerDb(db, "业务库", "mysql")
ContainerDb(queryDb, "查询视图库", "mysql/es/mongodb")


Rel(command, commandModel, "业务操作")
Rel(commandModel, db, "写业务")
Rel(commandModel, event, "业务事件")
Rel(event, queryDb, "更新视图库")

Rel(queryModel, query, "查询")
Rel(queryDb, queryModel, "查询")

Lay_R(command, query)
Lay_R(db, queryDb)
Lay_R(commandModel, queryModel)

Lay_D(query, queryModel)
Lay_D(event, queryDb)
Lay_D(event, db)
```


https://www.cnblogs.com/davenkin/p/cqrs-coding-practices.html


## 单元测试建议


### 引入 spock 

* 引入 groovy 的 [spock](http://spockframework.org/spock/docs/1.3/index.html) 单元测试
* 依赖数据库的单元测试，引入 [spock-dbuint](https://github.com/janbols/spock-dbunit)

https://juejin.im/post/5d8ad44a51882509615bc937

```

// 业务类
@Slf4j
@Service
@AllArgsConstructor
public class UserServiceImpl implements UserService {

    private final ApplicationEventPublisher eventPublisher;
    private final UserRepository userRepository;

    @Override
    @Transactional()
    public void addUser(UserEntity user) {
        user.setPassword("hello1");
        userRepository.save(user);
        eventPublisher.publishEvent(CreateUserEvent.builder()
                .id(user.getId())
                .username(user.getUsername())
                .password(user.getPassword())
                .build()
        );

    }

    @Override
    @Transactional(readOnly = true)
    public List<UserEntity> getAllUsers() {
        return userRepository.findAll();
    }
}

// 测试类


@ContextConfiguration(classes = [
        UserServiceImpl.class
])
class UserServiceImplTest extends Specification {

    @Autowired
    UserService userService;

    @SpringBean
    UserRepository userRepository = Mock()

    @SpringBean
    ApplicationListener listener = Mock()

    def "测试创建用户"() {
        given:
        def reqUser = new UserEntity(username: "hello")
        def resultUser = new UserEntity(username: reqUser.username, password: "hello1")
        def userEvent = new CreateUserEvent(username: resultUser.username, password: resultUser.password)
        when:
        userService.addUser(reqUser)
        then:
        1 * userRepository.save(resultUser)
        1 * listener.onApplicationEvent({ it.payload == userEvent })
    }
}



@SpringBootTest
class UserServiceDBTest extends Specification {

    @Autowired
    DataSource dataSource

    @Autowired
    UserService userService

    @DbUnit(
            configure = { IDatabaseTester it ->
                it.setUpOperation = DatabaseOperation.CLEAN_INSERT
                it.tearDownOperation = DatabaseOperation.TRUNCATE_TABLE
            },
            content = {
                user_entity(id: 1, username: "hello")
                user_entity(id: 2, username: "world")
            }
    )
    def "测试查询"() {
        given:
        def hello = userService.getAllUsers()

        expect:
        hello == [
                new UserEntity(id: 1, username: "hello"),
                new UserEntity(id: 2, username: "world")
        ]
    }
}

```

### 单元测试准则

https://github.com/yangyubo/zh-unit-testing-guidelines


## 开发规范


### 预防空指针


* 方法返回值为非 `java.util.Colleciton`、 `java.util.Map` 与原生类型，存在可能返回 `null` 时，方法返回签名需要为 `Optional<T>`

```
public User getUser(String id); // 反例
public Optional<User> getUser(String id); // 正例
```

* 方法返回值为 `java.util.Colleciton` 与原生类型时，禁止返回 `null`，需要返回 Collections.EMPTY_(LIST|MAP|SET)

```
// 反例
public List<String> list() {
    return null; 
}

// 正例
public List<String> list() {
    return Collections.EMPTY_LIST; 
}

```

* 处理不信任的返回值时，需要使用 `Optional.of(result)`

### Spring 注入

推荐使用构造方法注入
https://www.cnblogs.com/joemsu/p/7688307.html

```

@Service
@RequiredArgsConstructor
public class UserServiceImpl {
    private final UserRepository userRepository;
}

```


不推荐

```
@Service
public class UserServiceImpl {
    @Autowired
    private UserRepository userRepository;
}


@Service
public class UserServiceImpl {

    private UserRepository userRepository;

    @Autowired
    public void setUserRepository(UserRepository userRepository) {
        this.userRepository = userRepository;
    }
}

```

### 提升代码可阅读性

#### 模型转换使用自定义的 Converter

正例
```

UserDo userDo = userVo2DoConverter.convert(userVo);

// Conveter代码
interface Converter<S, D> {
    D convert(S source);
    default Optional<D> safeConvert(S source) {
        return Optional.ofNullable(convert(source));
    }
}

final class UserConverters {
    public static final Converter<UserVo, UserDo> userVo2DoConverter = new Converter<UserVo, UserDo>() {
        public UserDo convert(UserVo userVo) {
            return UserDo.builder()
                .username(userVo.getUserName())
                .build();
        }
    }
}

```

反例1
```
UserDo userDo = new UserDo();
userDo.setUsername(userVo.getUserName());
userDo.setUsername2(userVo.getUserName2());
userDo.setUsername3(userVo.getUserName3());
userDo.setUsername4(userVo.getUserName4());
userDo.setUsername5(userVo.getUserName5());
```

反例2
```
BeanUtils.copyBean(source, dest);
```

#### spring 的配置注入

推荐使用 `@ConfigurationProperties(prefix = "xxx.xxx")` 与 Bean 结合的方式，并使用 `spring-boot-configuration-processor` 自动生成 `spring-configuration-metadata.json` 文件，达到 idea 自动提示的效果。

```
@ConfigurationProperties(prefix = "spring.datasource")
public class DataSourceProperties {

	/**
	 * Name of the datasource. Default to "testdb" when using an embedded database.
	 */
	private String name;

	/**
	 * Whether to generate a random datasource name.
	 */
	private boolean generateUniqueName;

}

```


不建议使用 `@Value`

```
 @Value("${datasource.driverClassName}")
 private String driverClassName;
 @Value("${datasource.url}")
 private String url;
 @Value("${datasource.username}")
 private String username;
 @Value("${datasource.password}")
 private String password;
```

#### 其他

* public 方法放在 private 方法前面
* 重视 idea 的黄色警告




### Rest API 接口规范

#### 命名规范

* uri 要求是名词，表示资源路径
* 不用大写字母
* 单词之间使用 `-` 与下划线 `_`
* uri 中的名词表示资源集合，使用复数形式
* 当http方法不足以表示动作时，可在url最后加上以下划线开头的动词，表明动作(`/_login`)


#### 资源集合 vs 单个资源

URI表示资源的两种方式：资源集合、单个资源。


资源集合：

```
/zoos //所有动物园
/zoos/1/animals //id为1的动物园中的所有动物
```

单个资源：

```
/zoos/1 //id为1的动物园
/zoos/1;2;3 //id为1，2，3的动物园
```


#### 避免层级过深的 uri （重要）

`/` 在url中表达层级，用于按实体关联关系进行对象导航，一般根据id导航。

过深的导航容易导致url膨胀，不易维护，如 `GET /zoos/1/areas/3/animals/4`，尽量使用查询参数代替路径中的实体导航，如`GET /animals?zoo=1&area=3`；


需要分析出资源的关系，是**继承关系**还是**组合关系**

#### HTTP 方法

通过标准 HTTP 方法对资源CRUD：

GET：查询

```
GET /zoos
GET /zoos/1
GET /zoos/1/employees
```

POST：创建单个资源。POST一般向“资源集合”型uri发起

```
POST /animals  //新增动物
POST /zoos/1/employees //为id为1的动物园雇佣员工
```

PUT：更新单个资源（全量），客户端提供完整的更新后的资源。与之对应的是 PATCH，PATCH 负责部分更新，客户端提供要更新的那些字段。PUT/PATCH一般向“单个资源”型uri发起
```
PUT /animals/1
PUT /zoos/1
```

DELETE：删除
```
DELETE /zoos/1/employees/2
DELETE /zoos/1/employees/2;4;5
DELETE /zoos/1/animals  //删除id为1的动物园内的所有动物
```


参考: [Restful API 的设计规范](http://novoland.github.io/%E8%AE%BE%E8%AE%A1/2015/08/17/Restful%20API%20%E7%9A%84%E8%AE%BE%E8%AE%A1%E8%A7%84%E8%8C%83.html)


## 版本与文档规范


### 版本管理

* 版本序号定义三位，X.Y.Z-RELEASE
  * X 大版本号，对应史诗级迭代
  * Y 常规版本号，对应普通迭代
  * Z 修复优化版本号，对应 bug 修复等情况
* 利用 git 的 tag 进行代码的版本控制，tag 使用版本号名称，可利用自动化脚本发布，如 maven 的 maven-release-plugin  


### README 规范


```
## 项目简介

## 快速开发

## 


```


### RELEASE 文档规范


* 每次变更需要编写 `CHANGELOG.md`
* 每次变更需要把涉及到的 `DDL`、`DML` 记录在 ./db/migration)/`版本名称.sql` 文件里

#### CHANGELOG 格式

```

# CHANGELOG
=====

# [1.0.0-RELEASE](2020-04-01)

Features    :
* [新增骑手附近的商户功能](http://xxx.xxx.xxx/xxx/xxx)
* [新增骑手附近的商户功能](http://xxx.xxx.xxx/xxx/xxx)


Bugfixes:
* 修复了xxx场景下的空指针问题

SQL:
* [1.0.0-RELEASE.sql](./db/migration/1.0.0-RELEASE.sql)

```
